#include "matrixmath.h"
#include "sfm.h"
#include "obj.h"
#include <gtk/gtk.h>
#include "editor.h"
#include "editorobj.h"
#include "editorview.h"
#include "editorlist.h"
#include "config.h"


static void EditorObjListDestroyCB(gpointer data);

gpointer EditorObjListGetObj(editor_struct *e, gint row);
gint EditorObjListMatchByName(editor_struct *e, const gchar *name);

/* Object realizing. */
void EditorObjListRealizeAll(editor_struct *e, gboolean rerealize);

/* Object list IO. */
void EditorObjListUpdateRow(
        editor_struct *e, gint row, gpointer obj
);
void EditorObjListUpdateAll(editor_struct *e);

gint EditorObjListInsertRow(
        editor_struct *e, gint row, gpointer obj
);

/* Object values to widgets IO. */
void EditorObjListDoGetValues(
	editor_struct *e, gint obj_num, const gpointer obj
);
void EditorObjListDoApplyValues(editor_struct *e, gpointer obj);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Row data delete callback.
 */
static void EditorObjListDestroyCB(gpointer data)
{
	EditorObjDelete(EDITOR_OBJ(data));
}


/*
 *	Returns the object specified by the index row.
 */
gpointer EditorObjListGetObj(editor_struct *e, gint row)
{
	GtkCList *clist = (GtkCList *)((e != NULL) ? e->obj_clist : NULL);
	if(clist == NULL)
	    return(NULL);
	if((row < 0) || (row >= clist->rows))
	    return(NULL);
	else
	    return(gtk_clist_get_row_data(clist, row));
}

/*
 *	Returns the index row of the object who's name matches the
 *	given name.
 */
gint EditorObjListMatchByName(editor_struct *e, const gchar *name)
{
        gint i;
	GtkCList *clist = (GtkCList *)((e != NULL) ? e->obj_clist : NULL);
        if((clist == NULL) || (name == NULL))
            return(-1);

	if(*name == '\0')
	    return(-1);

        for(i = 0; i < clist->rows; i++)
        {
            editor_obj_struct *o = EDITOR_OBJ(
                gtk_clist_get_row_data(clist, i)
            );
            if((o != NULL) ? (o->name == NULL) : TRUE)
                continue;

	    if(*o->name == '\0')
		continue;

            if(!g_strcasecmp(o->name, name))
                return(i);
        }

        return(-1);
}


/*
 *      (Re)realizes all objects on the list, this updates the object's
 *      references (if any) and updates the cell label displayed for
 *      each object.
 */
void EditorObjListRealizeAll(editor_struct *e, gboolean rerealize)
{
	gint i;
	GtkCList *clist;

        if(e == NULL)
            return;

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	gtk_clist_freeze(clist);

	/* Iterate through all rows on the list. */
	for(i = 0; i < clist->rows; i++)
	{
	    editor_obj_struct *o = EDITOR_OBJ(
		gtk_clist_get_row_data(clist, i)
	    );
	    if(o == NULL)
		continue;

	    /* Begin realizing object values (as needed). */

	    /* Reference object. */
	    o->ref_obj_num = EditorObjListMatchByName(e, o->ref_obj_name);
	    /* If there is a referenced object then move this object to
	     * the correct location relative to the reference object.
	     */
	    if(o->ref_obj_num > -1)
	    {
		const editor_obj_struct *o2 = EDITOR_OBJ(
		    gtk_clist_get_row_data(clist, o->ref_obj_num)
		);
		if((o2 != o) && (o2 != NULL))
		{
		    gdouble a[4], r[4];
		    sar_position_struct *pos = &o->pos;
		    const sar_position_struct *ref_offset = &o->ref_offset;
		    const sar_direction_struct *ref_dir = &o->ref_dir;
		    const sar_position_struct *pos2 = &o2->pos;
		    const sar_direction_struct *dir2 = &o2->dir;

		    /* Calculate reference offset of o to o2. */
		    a[0] = ref_offset->x;
		    a[1] = ref_offset->y;
		    a[2] = ref_offset->z;
		    MatrixRotateHeading3(
			a, dir2->heading + ref_dir->heading, r
		    );
		    /* Move o to o2 and apply reference offset to o. */
		    memcpy(pos, pos2, sizeof(sar_position_struct));
		    pos->x += (float)r[0];
                    pos->y += (float)r[1];
		}
	    }

	    /* Update list cell value of object. */
	    EditorObjListUpdateRow(e, i, o);
	}

	gtk_clist_thaw(clist);
}


/*
 *	Updates the row with the given object values.
 */
void EditorObjListUpdateRow(
        editor_struct *e, gint row, gpointer obj
)
{
	gchar *name;
	GdkBitmap *mask;
	GdkPixmap *pixmap;
	GtkCList *clist;
	editor_obj_struct *o;


	if(e == NULL)
	    return;

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	if((row < 0) || (row >= clist->rows))
	    return;

	o = EDITOR_OBJ((obj != NULL) ?
	    obj :
	    gtk_clist_get_row_data(clist, row)
	);
	if(o == NULL)
	    return;

	/* Get name. */
	if(o->name != NULL)
	    name = STRDUP(o->name);
	else
	    name = g_strdup_printf("Object #%i", row);

	/* Get pixmap. */
	switch(o->type)
	{
	  case SAR_OBJ_TYPE_STATIC:
          case SAR_OBJ_TYPE_PREMODELED:
	    pixmap = e->list_object_pixmap;
	    mask = e->list_object_mask;
	    break;
          case SAR_OBJ_TYPE_GROUND:
            pixmap = e->list_ground_pixmap;
            mask = e->list_ground_mask;
            break;
          case SAR_OBJ_TYPE_HELIPAD:
            pixmap = e->list_helipad_pixmap;
            mask = e->list_helipad_mask;
            break;
          case SAR_OBJ_TYPE_RUNWAY:
            pixmap = e->list_runway_pixmap;
            mask = e->list_runway_mask;
            break;
          case SAR_OBJ_TYPE_HUMAN:
            pixmap = e->list_human_pixmap;
            mask = e->list_human_mask;
            break;
	  default:
	    pixmap = NULL;
	    mask = NULL;
	}

	/* Update cell. */
	if(pixmap != NULL)
	    gtk_clist_set_pixtext(
		clist, row, 0,
		name,
		SE_LIST_PIXMAP_TEXT_SPACING,
		pixmap,
		mask
	    );
	else
	    gtk_clist_set_text(
		clist, row, 0,
		name
	    );

	g_free(name);
}

/*
 *	Updates all rows with each row's object data.
 */
void EditorObjListUpdateAll(editor_struct *e)
{
	gint i;
	GtkCList *clist;

	if(e == NULL)
	    return;

        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

	gtk_clist_freeze(clist);

	for(i = 0; i < clist->rows; i++)
	    EditorObjListUpdateRow(e, i, NULL);

	gtk_clist_thaw(clist);
}

/*
 *	Inserts (or appends) a new row to the list.
 *
 *	The given object will be transfered to the new row's data
 *	so it should not be referenced again after this call.
 *
 *	Returns the new row index or -1 on error.
 */
gint EditorObjListInsertRow(
        editor_struct *e, gint row, gpointer obj
)
{
	gint i, new_row;
	gchar **strv;
        GtkCList *clist;
        editor_obj_struct *o = EDITOR_OBJ(obj);


        if((e == NULL) || (o == NULL))
	{
            EditorObjDelete(o);
            return(-1);
	}

        clist = (GtkCList *)e->obj_clist;
        if((clist != NULL) ? (clist->columns <= 0) : TRUE)
	{
	    EditorObjDelete(o);
            return(-1);
	}

	/* Allocate tempory cell values. */
	strv = (gchar **)g_malloc(clist->columns * sizeof(gchar *));
	for(i = 0; i < clist->columns; i++)
	    strv[i] = STRDUP("");

        gtk_clist_freeze(clist);

	/* Append row? */
	if(row < 0)
	    new_row = gtk_clist_append(clist, strv);
	else
	    new_row = gtk_clist_insert(clist, row, strv);

	/* New row created successfully? */
	if(new_row > -1)
	{
	    gtk_clist_set_row_data_full(
		clist, new_row, o, EditorObjListDestroyCB
	    );
	    EditorObjListUpdateRow(
		e, new_row, o
	    );
	}
	else
	{
	    EditorObjDelete(o);
	}

        gtk_clist_thaw(clist);

	/* Delete tempory cell values. */
        for(i = 0; i < clist->columns; i++)
            g_free(strv[i]);
	g_free(strv);

	return(new_row);
}


/*
 *	Gets values from object and sets them to the object widgets.
 *
 *	If the given obj is NULL, then the values in the object widgets
 *	will be reset/cleared.
 */
void EditorObjListDoGetValues(
	editor_struct *e, gint obj_num, const gpointer obj
)
{
	gint size_shape;
	gboolean polygon_offset;
	const gchar *vstr;
	gboolean vb;
	gfloat vf;
	gboolean is_alt;
	GtkAdjustment *adj;
	GtkWidget *w;
	const sar_contact_bounds_struct *cb;
	const editor_obj_struct *o = EDITOR_OBJ(obj);
        if(e == NULL)
            return;

	cb = (o != NULL) ? &o->cb : NULL;

        e->obj_freeze_count++;

#define SET_ADJ_VF						\
{ if(adj != NULL) {						\
 adj->value = vf;						\
 switch(EDITOR_UNITS(e))					\
 {								\
  case EDITOR_UNITS_METRIC:					\
   break;							\
  case EDITOR_UNITS_METRIC_ALT_FEET:				\
   if(is_alt)							\
    adj->value = (gfloat)SFMMetersToFeet(adj->value);		\
   break;							\
  case EDITOR_UNITS_ENGLISH:					\
   adj->value = (gfloat)SFMMetersToFeet(adj->value);		\
   break;							\
 }								\
 gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");	\
} }
#define SET_ADJ_VF_RAD_TO_DEG					\
{ if(adj != NULL) {						\
 adj->value = (gfloat)RADTODEG(vf);				\
 gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");	\
} }
#define SET_ENTRY						\
{ if((w != NULL) ? GTK_IS_ENTRY(w) : FALSE) {			\
 gtk_entry_set_text(GTK_ENTRY(w), (vstr != NULL) ? vstr : "");	\
} }
#define SET_TB_EMIT						\
{ if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE) {		\
 gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), vb);	\
} }
	/* Name. */
	w = e->obj_name_entry;
	if(w != NULL)
	{
	    gchar *buf;

	    if(o != NULL)
	    {
		if(o->name != NULL)
		    buf = STRDUP(o->name);
		else
		    buf = STRDUP("");
	    }
	    else
	    {
		buf = STRDUP("");
	    }
	    gtk_entry_set_text(GTK_ENTRY(w), buf);
	    g_free(buf);
	}

	/* Type. */
	EditorSetObjType(
	    e->obj_type_combo, (o != NULL) ? o->type : SAR_OBJ_TYPE_STATIC
	);

        /* Model file. */
        w = e->obj_model_entry;
        if(w != NULL)
        {
            gchar *buf;

            if(o != NULL)
            {
                if(o->model_file != NULL)
                    buf = STRDUP(o->model_file);
                else
                    buf = STRDUP("");
            }
            else
            {
                buf = STRDUP("");
            }
            gtk_entry_set_text(GTK_ENTRY(w), buf);
            g_free(buf);
        }

	/* X. */
	w = e->obj_x_spin;
	adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	is_alt = FALSE;
	vf = (o != NULL) ? (gfloat)o->pos.x : 0.0f;
	SET_ADJ_VF
	/* Y. */
        w = e->obj_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (o != NULL) ? (gfloat)o->pos.y : 0.0f;
        SET_ADJ_VF
	/* Z. */
        w = e->obj_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = TRUE;
        vf = (o != NULL) ? (gfloat)o->pos.z : 0.0f;
        SET_ADJ_VF

	/* Heading. */
        w = e->obj_heading_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->dir.heading : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG
	/* Pitch. */
        w = e->obj_pitch_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->dir.pitch : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG
	/* Bank. */
        w = e->obj_bank_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->dir.bank : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG

	/* Range. */
        w = e->obj_range_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (o != NULL) ? (gfloat)o->range : 0.0f;
        SET_ADJ_VF
	/* Range far. */
        w = e->obj_range_far_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (o != NULL) ? (gfloat)o->range_far : 0.0f;
        SET_ADJ_VF

	/* No depth test. */
	w = e->obj_no_depth_test_tb;
	vb = (gboolean)((o != NULL) ?
	    (o->flags & SAR_OBJ_FLAG_NO_DEPTH_TEST) : FALSE
	);
	SET_TB_EMIT
        /* Polygon offset. */
        w = e->obj_polygon_offset_tb;
	polygon_offset = vb = (gboolean)((o != NULL) ?
            (o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET) : FALSE
        );
        SET_TB_EMIT
        /* Polygon offset reverse. */
        w = e->obj_polygon_offset_reverse_tb;
        vb = (gboolean)((o != NULL) ?
            (o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_REVERSE) : FALSE
        );
        SET_TB_EMIT
	if(w != NULL)
	    gtk_widget_set_sensitive(w, polygon_offset);
        /* Polygon offset write depth. */
        w = e->obj_polygon_offset_write_depth_tb;
        vb = (gboolean)((o != NULL) ?
            (o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_WRITE_DEPTH) : FALSE
        );
        SET_TB_EMIT
        if(w != NULL)
            gtk_widget_set_sensitive(w, polygon_offset);

        /* Size shape. */
	size_shape = (cb != NULL) ? cb->contact_shape : SAR_CONTACT_SHAPE_SPHERICAL;
        EditorSetSizeShape(e->obj_size_shape_combo, size_shape);
        /* Size radius. */
        w = e->obj_size_radius_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (cb != NULL) ? (gfloat)cb->contact_radius : 0.0f;
        SET_ADJ_VF
        /* Size x min and max. */
        w = e->obj_size_x_min_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (cb != NULL) ? (gfloat)cb->contact_x_min : 0.0f;
        SET_ADJ_VF
        w = e->obj_size_x_max_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (cb != NULL) ? (gfloat)cb->contact_x_max : 0.0f;
        SET_ADJ_VF
        /* Size y min and max. */
        w = e->obj_size_y_min_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (cb != NULL) ? (gfloat)cb->contact_y_min : 0.0f;
        SET_ADJ_VF
        w = e->obj_size_y_max_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (cb != NULL) ? (gfloat)cb->contact_y_max : 0.0f;
        SET_ADJ_VF
        /* Size z min and max. */
	if(size_shape == SAR_CONTACT_SHAPE_CYLENDRICAL)
	{
	    w = e->obj_size_z_min_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            vf = (cb != NULL) ? (gfloat)cb->contact_h_min : 0.0f;
            SET_ADJ_VF
            w = e->obj_size_z_max_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            vf = (cb != NULL) ? (gfloat)cb->contact_h_max : 0.0f;
            SET_ADJ_VF
	}
	else
	{
            w = e->obj_size_z_min_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            vf = (cb != NULL) ? (gfloat)cb->contact_z_min : 0.0f;
            SET_ADJ_VF
            w = e->obj_size_z_max_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            vf = (cb != NULL) ? (gfloat)cb->contact_z_max : 0.0f;
            SET_ADJ_VF
	}

        /* Reference object name. */
        w = e->obj_ref_name_entry;
        if(w != NULL)
        {
            gchar *buf;

            if(o != NULL)
            {
                if(o->ref_obj_name != NULL)
                    buf = STRDUP(o->ref_obj_name);
                else
                    buf = STRDUP("");
            }
            else
            {
                buf = STRDUP("");
            }
            gtk_entry_set_text(GTK_ENTRY(w), buf);
            g_free(buf);
        }

	/* Reference offset position. */
        w = e->obj_ref_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (o != NULL) ? (gfloat)o->ref_offset.x : 0.0f;
        SET_ADJ_VF
        w = e->obj_ref_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        vf = (o != NULL) ? (gfloat)o->ref_offset.y : 0.0f;
        SET_ADJ_VF
        w = e->obj_ref_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = TRUE;
        vf = (o != NULL) ? (gfloat)o->ref_offset.z : 0.0f;
        SET_ADJ_VF

        /* Reference offset direction. */
        w = e->obj_ref_heading_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->ref_dir.heading : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG
        w = e->obj_ref_pitch_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->ref_dir.pitch : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG
        w = e->obj_ref_bank_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        vf = (gfloat)((o != NULL) ? o->ref_dir.bank : 0.0f * PI);
        SET_ADJ_VF_RAD_TO_DEG


        /* Helipad values. */
        w = e->obj_helipad_frame;
        if((o != NULL) ? (o->type == SAR_OBJ_TYPE_HELIPAD) : FALSE)
        {
            GtkEntry *entry;
	    gulong flags = o->helipad_flags;


            if(w != NULL)
                gtk_widget_show(w);

            w = e->obj_helipad_type_combo;
            entry = (GtkEntry *)((w != NULL) ? GTK_COMBO(w)->entry : NULL);
            if(entry != NULL)
                gtk_entry_set_text(
                    entry,
                    (o->helipad_type_name != NULL) ? o->helipad_type_name : ""
                );

            w = e->obj_helipad_edge_lighting_tb;
            vb = (gboolean)(flags & SAR_HELIPAD_FLAG_EDGE_LIGHTING);
            SET_TB_EMIT
            w = e->obj_helipad_fuel_tb;
            vb = (gboolean)(flags & SAR_HELIPAD_FLAG_FUEL);
            SET_TB_EMIT
            w = e->obj_helipad_repair_tb;
            vb = (gboolean)(flags & SAR_HELIPAD_FLAG_REPAIR);
            SET_TB_EMIT
            w = e->obj_helipad_dropoff_tb;
            vb = (gboolean)(flags & SAR_HELIPAD_FLAG_DROPOFF);
            SET_TB_EMIT
            w = e->obj_helipad_restart_point_tb;
            vb = (gboolean)(flags & SAR_HELIPAD_FLAG_RESTART_POINT);
            SET_TB_EMIT

	    w = e->obj_helipad_length_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->helipad_length;
            SET_ADJ_VF
            w = e->obj_helipad_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->helipad_width;
            SET_ADJ_VF
            w = e->obj_helipad_recession_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            vf = o->helipad_recession;
            SET_ADJ_VF

            w = e->obj_helipad_label_entry;
            vstr = o->helipad_label;
            SET_ENTRY
        }
        else
        {
            if(w != NULL)
                gtk_widget_hide(w);
        }

        /* Runway values. */
        w = e->obj_runway_frame;
        if((o != NULL) ? (o->type == SAR_OBJ_TYPE_RUNWAY) : FALSE)
        {
            gulong flags = o->runway_flags;
	    gulong al_flags;

            if(w != NULL)
                gtk_widget_show(w);

	    /* Surface. */
	    EditorSetRunwaySurface(
		e->obj_runway_surface_combo, o->runway_surface
	    );

	    /* Length and width. */
            w = e->obj_runway_length_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->runway_length;
            SET_ADJ_VF
            w = e->obj_runway_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->runway_width;
            SET_ADJ_VF

	    /* Dashes. */
            w = e->obj_runway_dashes_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = (gfloat)o->runway_dashes;
            SET_ADJ_VF

            /* Edge light spacing. */
            w = e->obj_runway_edge_light_spacing_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->runway_edge_light_spacing;
            SET_ADJ_VF

	    /* Markings. */
	    w = e->obj_runway_thresholds_tb;
	    vb = (flags & SAR_RUNWAY_FLAG_THRESHOLDS) ? TRUE : FALSE;
	    SET_TB_EMIT
            w = e->obj_runway_borders_tb;
            vb = (flags & SAR_RUNWAY_FLAG_BORDERS) ? TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_td_markers_tb;
            vb = (flags & SAR_RUNWAY_FLAG_TD_MARKERS) ? TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_midway_markers_tb;
            vb = (flags & SAR_RUNWAY_FLAG_MIDWAY_MARKERS) ? TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_north_gs_tb;
            vb = (flags & SAR_RUNWAY_FLAG_NORTH_GS) ? TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_south_gs_tb;
            vb = (flags & SAR_RUNWAY_FLAG_SOUTH_GS) ? TRUE : FALSE;
            SET_TB_EMIT

	    /* End labels. */
            w = e->obj_runway_north_label_entry;
            vstr = o->runway_north_label;
            SET_ENTRY
            w = e->obj_runway_south_label_entry;
            vstr = o->runway_south_label;
            SET_ENTRY

            /* Displaced threshold. */
            w = e->obj_runway_north_displaced_threshold_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->runway_north_displaced_threshold;
            SET_ADJ_VF
            w = e->obj_runway_south_displaced_threshold_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            vf = o->runway_south_displaced_threshold;
            SET_ADJ_VF

	    /* North end approach lighting. */
	    al_flags = o->runway_north_lighting_flags;
	    w = e->obj_runway_north_lighting_end_tb;
	    vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_END) ?
		TRUE : FALSE;
	    SET_TB_EMIT
            w = e->obj_runway_north_lighting_tracer_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_TRACER) ?
		TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_north_lighting_align_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_ALIGN) ?
                TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_north_lighting_ils_glide_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_ILS_GLIDE) ?
                TRUE : FALSE;
            SET_TB_EMIT

            /* South end approach lighting. */
            al_flags = o->runway_south_lighting_flags;
            w = e->obj_runway_south_lighting_end_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_END) ?
                TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_south_lighting_tracer_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_TRACER) ?
                TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_south_lighting_align_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_ALIGN) ?
                TRUE : FALSE;
            SET_TB_EMIT
            w = e->obj_runway_south_lighting_ils_glide_tb;
            vb = (al_flags & SAR_RUNWAY_APPROACH_LIGHTING_ILS_GLIDE) ?
                TRUE : FALSE;
            SET_TB_EMIT
        }
        else
        {
            if(w != NULL)
                gtk_widget_hide(w);
        }


	/* Human values. */
	w = e->obj_human_frame;
	if((o != NULL) ? (o->type == SAR_OBJ_TYPE_HUMAN) : FALSE)
	{
	    GtkEntry *entry;
	    gulong flags = o->human_flags;

	    if(w != NULL)
		gtk_widget_show(w);

	    w = e->obj_human_type_combo;
	    entry = (GtkEntry *)((w != NULL) ? GTK_COMBO(w)->entry : NULL);
	    if(entry != NULL)
		gtk_entry_set_text(
		    entry,
		    (o->human_type_name != NULL) ? o->human_type_name : ""
		);

	    w = e->obj_human_need_rescue_tb;
	    vb = (gboolean)(flags & SAR_HUMAN_FLAG_NEED_RESCUE);
	    SET_TB_EMIT

            w = e->obj_human_sit_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_SIT);
            SET_TB_EMIT
            w = e->obj_human_sit_up_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_SIT_UP);
            SET_TB_EMIT
            w = e->obj_human_sit_down_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_SIT_DOWN);
            SET_TB_EMIT
            w = e->obj_human_lying_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_LYING);
            SET_TB_EMIT

            w = e->obj_human_alert_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_ALERT);
            SET_TB_EMIT
            w = e->obj_human_aware_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_AWARE);
            SET_TB_EMIT
            w = e->obj_human_in_water_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_IN_WATER);
            SET_TB_EMIT
            w = e->obj_human_on_streatcher_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_ON_STREATCHER);
            SET_TB_EMIT

#if 0
            w = e->obj_human_run_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_RUN);
            SET_TB_EMIT
#endif
            w = e->obj_human_run_towards_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_RUN_TOWARDS);
            SET_TB_EMIT
            w = e->obj_human_run_away_tb;
            vb = (gboolean)(flags & SAR_HUMAN_FLAG_RUN_AWAY);
            SET_TB_EMIT

	    w = e->obj_human_message_enter_entry;
	    vstr = o->human_message_enter;
	    SET_ENTRY
	}
	else
	{
            if(w != NULL)
                gtk_widget_hide(w);
	}

        /* Preset model values. */
        w = e->obj_preset_model_frame;
        if((o != NULL) ? (o->type == SAR_OBJ_TYPE_PREMODELED) : FALSE)
        {
            GtkEntry *entry;


            if(w != NULL)
                gtk_widget_show(w);

            w = e->obj_preset_model_type_combo;
            entry = (GtkEntry *)((w != NULL) ? GTK_COMBO(w)->entry : NULL);
            if(entry != NULL)
                gtk_entry_set_text(
                    entry,
                    (o->premodeled_type_name != NULL) ?
			o->premodeled_type_name : ""
                );

            w = e->obj_preset_model_values_entry;
            vstr = o->premodeled_value;
            SET_ENTRY
        }
        else
        {
            if(w != NULL)
                gtk_widget_hide(w);
        }

#undef SET_TB_EMIT
#undef SET_ENTRY
#undef SET_ADJ_VF_RAD_TO_DEG
#undef SET_ADJ_VF

	e->obj_freeze_count--;

	EditorUpdateMenus(e);
}

/*
 *	Sets values from the object widgets to the object.
 *
 *	If the obj is NULL then the last selected object will be
 *	used (if any).
 */
void EditorObjListDoApplyValues(editor_struct *e, gpointer obj)
{
	gboolean new_changes = FALSE;
	const gchar *obj_ref_name;
	gulong flag;
	gboolean vb;
	gint vi;
	gfloat vf;
	gboolean is_alt;
	GtkAdjustment *adj;
	GtkWidget *w;
	GtkCList *clist;
	sar_contact_bounds_struct *cb;
	editor_obj_struct *o = EDITOR_OBJ(obj);
	if(e == NULL)
	    return;

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	if(o == NULL)
	{
	    GList *glist = clist->selection_end;
	    gint row = (glist != NULL) ? (gint)glist->data : -1;
	    if(row < 0)
		return;

	    o = EDITOR_OBJ(gtk_clist_get_row_data(clist, row));
	}
	if(o == NULL)
	    return;

        cb = &o->cb;

	/* Get reference object name (if any). */
        w = e->obj_ref_name_entry;
	obj_ref_name = (w != NULL) ?
	    gtk_entry_get_text(GTK_ENTRY(w)) : NULL;

#define GET_ADJ_VF			\
{ if(adj != NULL) {			\
 vf = adj->value;			\
 switch(EDITOR_UNITS(e))		\
 {					\
  case EDITOR_UNITS_METRIC:		\
   break;				\
  case EDITOR_UNITS_METRIC_ALT_FEET:	\
   if(is_alt)				\
    vf = (gfloat)SFMFeetToMeters(vf);	\
   break;				\
  case EDITOR_UNITS_ENGLISH:		\
   vf = (gfloat)SFMFeetToMeters(vf);	\
   break;				\
 }					\
} else {				\
 vf = 0.0f;				\
} }
#define GET_ADJ_VF_DEG_TO_RAD		\
{ if(adj != NULL) {			\
 vf = (gfloat)DEGTORAD(adj->value);	\
 while(vf >= (gfloat)(2.0 * PI))	\
  vf -= (gfloat)(2.0 * PI);		\
 while(vf < (gfloat)(0.0 * PI))		\
  vf += (gfloat)(2.0 * PI);		\
} else {				\
 vf = (gfloat)(0.0 * PI);		\
} }
#define GET_TB						\
{ if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE) {	\
 vb = gtk_toggle_button_get_active(			\
  GTK_TOGGLE_BUTTON(w)					\
 );							\
} else {						\
 vb = FALSE;						\
} }
	/* Name. */
	w = e->obj_name_entry;
	if(w != NULL)
	{
	    const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
	    if(IS_STRING_EMPTY(s))
	    {
                if(o->name != NULL)
                {
                    g_free(o->name);
                    o->name = NULL;
                    new_changes = TRUE;
                }
            }
	    else if((o->name != NULL) ?
		strcmp(o->name, s) : TRUE
	    )
	    {
		g_free(o->name);
		o->name = STRDUP(s);
		new_changes = TRUE;
	    }
	}

        /* Type. */
	vi = EditorGetObjType(e->obj_type_combo);
	if(vi != o->type)
	{
	    o->type = vi;
	    new_changes = TRUE;
	}

        /* Model file. */
        w = e->obj_model_entry;
        if(w != NULL)
        {
            const gchar *s = gtk_entry_get_text(GTK_ENTRY(w));
            if(IS_STRING_EMPTY(s))
            {
                if(o->model_file != NULL)
                {
                    g_free(o->model_file);
                    o->model_file = NULL;
                    new_changes = TRUE;
                }
            }
	    else if((o->model_file != NULL) ?
		strcmp(o->model_file, s) : TRUE
	    )
	    {
                g_free(o->model_file);
                o->model_file = STRDUP(s);
                new_changes = TRUE;
            }
        }

	/* Set position only if there is no reference object specified. */
	if(IS_STRING_EMPTY(obj_ref_name))
	{
	    /* Position X. */
	    w = e->obj_x_spin;
	    adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	    is_alt = FALSE;
	    GET_ADJ_VF
	    if((gfloat)o->pos.x != vf)
	    {
		o->pos.x = vf;
		new_changes = TRUE;
	    }
	    /* Position Y. */
	    w = e->obj_y_spin;
	    adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	    is_alt = FALSE;
	    GET_ADJ_VF
	    if((gfloat)o->pos.y != vf)
	    {
		o->pos.y = vf;
		new_changes = TRUE;
	    }
	    /* Position Z. */
	    w = e->obj_z_spin;
	    adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	    is_alt = TRUE;
	    GET_ADJ_VF
	    if((gfloat)o->pos.z != vf)
	    {
		o->pos.z = vf;
		new_changes = TRUE;
	    }
	}

        /* Heading. */
        w = e->obj_heading_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->dir.heading != vf)
	{
	    o->dir.heading = vf;
	    new_changes = TRUE;
	}
        /* Pitch. */
        w = e->obj_pitch_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->dir.pitch != vf)
        {
            o->dir.pitch = vf;
            new_changes = TRUE;
        }
        /* Bank. */
        w = e->obj_bank_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->dir.bank != vf)
        {
            o->dir.bank = vf;
            new_changes = TRUE;
        }

        /* Range. */
	w = e->obj_range_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        GET_ADJ_VF
        if((gfloat)o->range != vf)
        {
            o->range = vf;
            new_changes = TRUE;
        }
        /* Range far. */
        w = e->obj_range_far_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        GET_ADJ_VF
        if((gfloat)o->range_far != vf)
        {
            o->range_far = vf;
            new_changes = TRUE;
        }

        /* No depth test. */
        w = e->obj_no_depth_test_tb;
	GET_TB
	flag = SAR_OBJ_FLAG_NO_DEPTH_TEST;
	if(vb && !(o->flags & flag))
	{
	    o->flags |= flag;
	    new_changes = TRUE;
	}
	else if(!vb && (o->flags & flag))
	{
	    o->flags &= ~flag;
	    new_changes = TRUE;
	}
        /* Polygon offset. */
        w = e->obj_polygon_offset_tb;
        GET_TB
	flag = SAR_OBJ_FLAG_POLYGON_OFFSET;
        if(vb && !(o->flags & flag))
	{
	    o->flags |= flag;
	    new_changes = TRUE;
	}
	else if(!vb && (o->flags & flag))
	{
	    o->flags &= ~flag;
	    new_changes = TRUE;
	}
        /* Polygon offset reverse. */
        w = e->obj_polygon_offset_reverse_tb;
        GET_TB
	flag = SAR_OBJ_FLAG_POLYGON_OFFSET_REVERSE;
        if(vb && !(o->flags & flag))
	{
	    o->flags |= flag;
	    new_changes = TRUE;
	}
	else if(!vb && (o->flags & flag))
	{
	    o->flags &= ~flag;
	    new_changes = TRUE;
	}
	/* Polygon offset write depth. */
        w = e->obj_polygon_offset_write_depth_tb;
        GET_TB
	flag = SAR_OBJ_FLAG_POLYGON_OFFSET_WRITE_DEPTH;
        if(vb && !(o->flags & flag))
	{
	    o->flags |= flag;
	    new_changes = TRUE;
	}
	else if(!vb && (o->flags & flag))
	{
	    o->flags &= ~flag;
	    new_changes = TRUE;
	}

	/* Size. */
	if(cb != NULL)
	{
	    gint size_shape;

	    /* Shape. */
	    size_shape = EditorGetSizeShape(e->obj_size_shape_combo);
	    if(size_shape != cb->contact_shape)
	    {
		cb->contact_shape = size_shape;
		new_changes = TRUE;
	    }
	    switch(size_shape)
	    {
	      case SAR_CONTACT_SHAPE_SPHERICAL:
		/* Radius. */
		w = e->obj_size_radius_spin;
		adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
		is_alt = FALSE;
		GET_ADJ_VF
		if((gfloat)cb->contact_radius != vf)
		{
		    cb->contact_radius = vf;
		    new_changes = TRUE;
		}
		break;

              case SAR_CONTACT_SHAPE_CYLENDRICAL:
                /* Radius. */
                w = e->obj_size_radius_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = FALSE;
                GET_ADJ_VF
                if((gfloat)cb->contact_radius != vf)
                {
                    cb->contact_radius = vf;
                    new_changes = TRUE;
                }
		/* Height min. */
                w = e->obj_size_z_min_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = TRUE;
                GET_ADJ_VF
                if((gfloat)cb->contact_h_min != vf)
                {
                    cb->contact_h_min = vf;
                    new_changes = TRUE;
                }
                /* Height max. */
                w = e->obj_size_z_max_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = TRUE;
                GET_ADJ_VF
                if((gfloat)cb->contact_h_max != vf)
                {
                    cb->contact_h_max = vf;
                    new_changes = TRUE;
                }
                break;

              case SAR_CONTACT_SHAPE_RECTANGULAR:
                /* X min and max. */
                w = e->obj_size_x_min_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = FALSE;
                GET_ADJ_VF
                if((gfloat)cb->contact_x_min != vf)
                {
                    cb->contact_x_min = vf;
                    new_changes = TRUE;
                }
                w = e->obj_size_x_max_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = FALSE;
                GET_ADJ_VF
                if((gfloat)cb->contact_x_max != vf)
                {
                    cb->contact_x_max = vf;
                    new_changes = TRUE;
                }
                /* Y min and max. */
                w = e->obj_size_y_min_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = FALSE;
                GET_ADJ_VF
                if((gfloat)cb->contact_y_min != vf)
                {
                    cb->contact_y_min = vf;
                    new_changes = TRUE;
                }
                w = e->obj_size_y_max_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = FALSE;
                GET_ADJ_VF
                if((gfloat)cb->contact_y_max != vf)
                {
                    cb->contact_y_max = vf;
                    new_changes = TRUE;
                }
                /* Z min and max. */
                w = e->obj_size_z_min_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = TRUE;
                GET_ADJ_VF
                if((gfloat)cb->contact_z_min != vf)
                {
                    cb->contact_z_min = vf;
                    new_changes = TRUE;
                }
                w = e->obj_size_z_max_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                is_alt = TRUE;
                GET_ADJ_VF
                if((gfloat)cb->contact_z_max != vf)
                {
                    cb->contact_z_max = vf;
                    new_changes = TRUE;
                }
		break;
	    }
	}

	/* Reference object name. */
	if(IS_STRING_EMPTY(obj_ref_name))
	{
	    if(o->ref_obj_name != NULL)
	    {
		g_free(o->ref_obj_name);
		o->ref_obj_name = NULL;
		new_changes = TRUE;
	    }
	}
	else if((o->ref_obj_name != NULL) ?
	    strcmp(o->ref_obj_name, obj_ref_name) : TRUE
	)
	{
	    g_free(o->ref_obj_name);
	    o->ref_obj_name = STRDUP(obj_ref_name);
	    new_changes = TRUE;
	}

        /* Reference offset position. */
        w = e->obj_ref_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        GET_ADJ_VF
        if((gfloat)o->ref_offset.x != vf)
        {
            o->ref_offset.x = vf;
            new_changes = TRUE;
        }
        w = e->obj_ref_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        GET_ADJ_VF
        if((gfloat)o->ref_offset.y != vf)
        {
            o->ref_offset.y = vf;
            new_changes = TRUE;
        }
        w = e->obj_ref_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = TRUE;
        GET_ADJ_VF
        if((gfloat)o->ref_offset.z != vf)
        {
            o->ref_offset.z = vf;
            new_changes = TRUE;
        }

	/* Reference offset direction. */
        w = e->obj_ref_heading_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->ref_dir.heading != vf)
        {
            o->ref_dir.heading = vf;
            new_changes = TRUE;
        }
        w = e->obj_ref_pitch_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->ref_dir.pitch != vf)
        {
            o->ref_dir.pitch = vf;
            new_changes = TRUE;
        }
        w = e->obj_ref_bank_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        GET_ADJ_VF_DEG_TO_RAD
        if((gfloat)o->ref_dir.bank != vf)
        {
            o->ref_dir.bank = vf;
            new_changes = TRUE;
        }


        /* Helipad values. */
        if(o->type == SAR_OBJ_TYPE_HELIPAD)
        {
            const gchar *s;
            GtkEntry *entry;

            /* Helipad type. */
            w = e->obj_helipad_type_combo;
            if(w != NULL)
            {
                entry = (GtkEntry *)((w != NULL) ?
                    GTK_COMBO(w)->entry : NULL
                );
                s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
                if(!IS_STRING_EMPTY(s))
                {
                    if(IS_STRING_EMPTY(o->helipad_type_name))
                    {
                        g_free(o->helipad_type_name);
                        o->helipad_type_name = STRDUP(s);
                        new_changes = TRUE;
                    }
                    else if(g_strcasecmp(s, o->helipad_type_name))
                    {
                        g_free(o->helipad_type_name);
                        o->helipad_type_name = STRDUP(s);
                        new_changes = TRUE;
                    }
		}
	    }

#define DO_SET_HELIPAD_FLAG			\
{ if(vb && !(o->helipad_flags & flag))		\
  {						\
    o->helipad_flags |= flag;			\
    new_changes = TRUE;				\
  }						\
  else if(!vb && (o->helipad_flags & flag))	\
  {						\
    o->helipad_flags &= ~flag;			\
    new_changes = TRUE;				\
  }						\
}

            w = e->obj_helipad_edge_lighting_tb;
            GET_TB
            flag = SAR_HELIPAD_FLAG_EDGE_LIGHTING;
            DO_SET_HELIPAD_FLAG
            w = e->obj_helipad_fuel_tb;
            GET_TB
            flag = SAR_HELIPAD_FLAG_FUEL;
            DO_SET_HELIPAD_FLAG
            w = e->obj_helipad_repair_tb;
            GET_TB
            flag = SAR_HELIPAD_FLAG_REPAIR;
            DO_SET_HELIPAD_FLAG
            w = e->obj_helipad_dropoff_tb;
            GET_TB
            flag = SAR_HELIPAD_FLAG_DROPOFF;
            DO_SET_HELIPAD_FLAG
            w = e->obj_helipad_restart_point_tb;
            GET_TB
            flag = SAR_HELIPAD_FLAG_RESTART_POINT;
            DO_SET_HELIPAD_FLAG

#undef DO_SET_HELIPAD_FLAG

	    /* Length, width, and recession. */
	    w = e->obj_helipad_length_spin;
	    adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	    is_alt = FALSE;
	    GET_ADJ_VF
	    if((gfloat)o->helipad_length != vf)
	    {
		o->helipad_length = vf;
		new_changes = TRUE;
	    }
            w = e->obj_helipad_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if((gfloat)o->helipad_width != vf)
            {
                o->helipad_width = vf;
                new_changes = TRUE;
            }
            w = e->obj_helipad_recession_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = TRUE;
            GET_ADJ_VF
            if((gfloat)o->helipad_recession != vf)
            {
                o->helipad_recession = vf;
                new_changes = TRUE;
            }

	    /* Helipad label. */
            entry = (GtkEntry *)e->obj_helipad_label_entry;
            s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
            if(IS_STRING_EMPTY(s))
            {
                if(o->helipad_label != NULL)
                {
                    g_free(o->helipad_label);
                    o->helipad_label = NULL;
                    new_changes = TRUE;
                }
            }
            else if((o->helipad_label != NULL) ?
                strcmp(s, o->helipad_label) : TRUE
            )
            {
                g_free(o->helipad_label);
                o->helipad_label = STRDUP(s);
                new_changes = TRUE;
            }
	}
        /* Runway values. */
        else if(o->type == SAR_OBJ_TYPE_RUNWAY)
        {
            const gchar *s;
            GtkEntry *entry;

	    /* Surface. */
            vi = EditorGetRunwaySurface(e->obj_runway_surface_combo);
	    if(o->runway_surface != vi)
	    {
		o->runway_surface = vi;
		new_changes = TRUE;
            }

            /* Length and width. */
            w = e->obj_runway_length_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if((gfloat)o->runway_length != vf)
            {
                o->runway_length = vf;
                new_changes = TRUE;
            }
            w = e->obj_runway_width_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if((gfloat)o->runway_width != vf)
            {
                o->runway_width = vf;
                new_changes = TRUE;
            }

	    /* Dashes. */
            w = e->obj_runway_dashes_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if(o->runway_dashes != (gint)vf)
            {
                o->runway_dashes = (gint)vf;
                new_changes = TRUE;
            }

            /* Edge light spacing. */
            w = e->obj_runway_edge_light_spacing_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if(o->runway_edge_light_spacing != vf)
            {
                o->runway_edge_light_spacing = vf;
                new_changes = TRUE;
            }

#define DO_SET_RUNWAY_FLAG			\
{ if(vb && !(o->runway_flags & flag))		\
  {                                             \
    o->runway_flags |= flag;			\
    new_changes = TRUE;                         \
  }                                             \
  else if(!vb && (o->runway_flags & flag))	\
  {                                             \
    o->runway_flags &= ~flag;			\
    new_changes = TRUE;                         \
  }                                             \
}
            w = e->obj_runway_thresholds_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_THRESHOLDS;
            DO_SET_RUNWAY_FLAG

            w = e->obj_runway_borders_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_BORDERS;
            DO_SET_RUNWAY_FLAG

	    w = e->obj_runway_td_markers_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_TD_MARKERS;
            DO_SET_RUNWAY_FLAG

            w = e->obj_runway_midway_markers_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_MIDWAY_MARKERS;
            DO_SET_RUNWAY_FLAG

            w = e->obj_runway_north_gs_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_NORTH_GS;
            DO_SET_RUNWAY_FLAG

            w = e->obj_runway_south_gs_tb;
            GET_TB
            flag = SAR_RUNWAY_FLAG_SOUTH_GS;
            DO_SET_RUNWAY_FLAG
#undef DO_SET_RUNWAY_FLAG

            /* North label. */
            entry = (GtkEntry *)e->obj_runway_north_label_entry;
            s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
            if(IS_STRING_EMPTY(s))
            {
                if(o->runway_north_label != NULL)
                {
                    g_free(o->runway_north_label);
                    o->runway_north_label = NULL;
                    new_changes = TRUE;
                }
            }
            else if((o->runway_north_label != NULL) ?
                strcmp(s, o->runway_north_label) : TRUE
            )
            {
                g_free(o->runway_north_label);
                o->runway_north_label = STRDUP(s);
                new_changes = TRUE;
            }
            /* South label. */
            entry = (GtkEntry *)e->obj_runway_south_label_entry;
            s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
            if(IS_STRING_EMPTY(s))
            {
                if(o->runway_south_label != NULL)
                {
                    g_free(o->runway_south_label);
                    o->runway_south_label = NULL;
                    new_changes = TRUE;
                }
            }
            else if((o->runway_south_label != NULL) ?
                strcmp(s, o->runway_south_label) : TRUE
            )
            {
                g_free(o->runway_south_label);
                o->runway_south_label = STRDUP(s);
                new_changes = TRUE;
            }

            /* Displaced thresholds. */
            w = e->obj_runway_north_displaced_threshold_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if(o->runway_north_displaced_threshold != (gfloat)vf)
            {
                o->runway_north_displaced_threshold = (gfloat)vf;
                new_changes = TRUE;
            }
            w = e->obj_runway_south_displaced_threshold_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            is_alt = FALSE;
            GET_ADJ_VF
            if(o->runway_south_displaced_threshold != (gfloat)vf)
            {
                o->runway_south_displaced_threshold = (gfloat)vf;
                new_changes = TRUE;
            }

#define DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG	\
{ if(vb && !(o->runway_north_lighting_flags & flag))	\
  {							\
    o->runway_north_lighting_flags |= flag;		\
    new_changes = TRUE;					\
  }							\
  else if(!vb && (o->runway_north_lighting_flags & flag)) \
  {							\
    o->runway_north_lighting_flags &= ~flag;		\
    new_changes = TRUE;					\
  }							\
}
            w = e->obj_runway_north_lighting_end_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_END;
            DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_north_lighting_tracer_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_TRACER;
            DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_north_lighting_align_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_ALIGN;
            DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_north_lighting_ils_glide_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_ILS_GLIDE;
            DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG

#undef DO_SET_RUNWAY_NORTH_APPROACH_LIGHTING_FLAG
#define DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG      \
{ if(vb && !(o->runway_south_lighting_flags & flag))	\
  {                                                     \
    o->runway_south_lighting_flags |= flag;             \
    new_changes = TRUE;                                 \
  }                                                     \
  else if(!vb && (o->runway_south_lighting_flags & flag)) \
  {                                                     \
    o->runway_south_lighting_flags &= ~flag;            \
    new_changes = TRUE;                                 \
  }                                                     \
}
            w = e->obj_runway_south_lighting_end_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_END;
            DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_south_lighting_tracer_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_TRACER;
            DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_south_lighting_align_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_ALIGN;
            DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG

            w = e->obj_runway_south_lighting_ils_glide_tb;
            GET_TB
            flag = SAR_RUNWAY_APPROACH_LIGHTING_ILS_GLIDE;
            DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG

#undef DO_SET_RUNWAY_SOUTH_APPROACH_LIGHTING_FLAG

	}
	/* Human values. */
	else if(o->type == SAR_OBJ_TYPE_HUMAN)
	{
	    const gchar *s;
	    GtkEntry *entry;

	    /* Human type. */
	    w = e->obj_human_type_combo;
	    if(w != NULL)
	    {
		entry = (GtkEntry *)((w != NULL) ?
		    GTK_COMBO(w)->entry : NULL
		);
		s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
		if(!IS_STRING_EMPTY(s))
		{
		    if(IS_STRING_EMPTY(o->human_type_name))
		    {
                        g_free(o->human_type_name);
                        o->human_type_name = STRDUP(s);
			new_changes = TRUE;
		    }
		    else if(g_strcasecmp(s, o->human_type_name))
		    {
			g_free(o->human_type_name);
			o->human_type_name = STRDUP(s);
			new_changes = TRUE;
		    }
		}
	    }

#define DO_SET_HUMAN_FLAG			\
{ if(vb && !(o->human_flags & flag))		\
  {						\
    o->human_flags |= flag;			\
    new_changes = TRUE;				\
  }						\
  else if(!vb && (o->human_flags & flag))	\
  {						\
    o->human_flags &= ~flag;			\
    new_changes = TRUE;				\
  }						\
}

	    w = e->obj_human_need_rescue_tb;
	    GET_TB
	    flag = SAR_HUMAN_FLAG_NEED_RESCUE;
	    DO_SET_HUMAN_FLAG

            w = e->obj_human_sit_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_SIT;
	    DO_SET_HUMAN_FLAG
            w = e->obj_human_sit_up_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_SIT_UP;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_sit_down_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_SIT_DOWN;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_lying_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_LYING;
            DO_SET_HUMAN_FLAG

            w = e->obj_human_alert_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_ALERT;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_aware_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_AWARE;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_in_water_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_IN_WATER;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_on_streatcher_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_ON_STREATCHER;
            DO_SET_HUMAN_FLAG

#if 0
            w = e->obj_human_run_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_RUN;
            DO_SET_HUMAN_FLAG
#endif
            w = e->obj_human_run_towards_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_RUN_TOWARDS;
            DO_SET_HUMAN_FLAG
            w = e->obj_human_run_away_tb;
            GET_TB
            flag = SAR_HUMAN_FLAG_RUN_AWAY;
            DO_SET_HUMAN_FLAG

#undef DO_SET_HUMAN_FLAG

	    entry = (GtkEntry *)e->obj_human_message_enter_entry;
	    s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
	    if(IS_STRING_EMPTY(s))
	    {
		if(o->human_message_enter != NULL)
		{
		    g_free(o->human_message_enter);
		    o->human_message_enter = NULL;
		    new_changes = TRUE;
		}
	    }
	    else if((o->human_message_enter != NULL) ?
		strcmp(s, o->human_message_enter) : TRUE
	    )
	    {
		g_free(o->human_message_enter);
		o->human_message_enter = STRDUP(s);
		new_changes = TRUE;
	    }
	}
        /* Preset model values. */
        else if(o->type == SAR_OBJ_TYPE_PREMODELED)
        {
            const gchar *s;
            GtkEntry *entry;

            /* Preset model type. */
            w = e->obj_preset_model_type_combo;
            if(w != NULL)
            {
                entry = (GtkEntry *)((w != NULL) ?
                    GTK_COMBO(w)->entry : NULL
                );
                s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
                if(!IS_STRING_EMPTY(s))
                {
                    if(IS_STRING_EMPTY(o->premodeled_type_name))
                    {
                        g_free(o->premodeled_type_name);
                        o->premodeled_type_name = STRDUP(s);
                        new_changes = TRUE;
                    }
                    else if(g_strcasecmp(s, o->premodeled_type_name))
                    {
                        g_free(o->premodeled_type_name);
                        o->premodeled_type_name = STRDUP(s);
                        new_changes = TRUE;
                    }
                }
            }

            entry = (GtkEntry *)e->obj_preset_model_values_entry;
            s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
            if(IS_STRING_EMPTY(s))
            {
                if(o->premodeled_value != NULL)
                {
                    g_free(o->premodeled_value);
                    o->premodeled_value = NULL;
                    new_changes = TRUE;
                }
            }
            else if((o->premodeled_value != NULL) ?
                strcmp(s, o->premodeled_value) : TRUE
            )
            {
                g_free(o->premodeled_value);
                o->premodeled_value = STRDUP(s);
                new_changes = TRUE;
            }
	}


#undef GET_TB
#undef GET_ADJ_VF_DEG_TO_RAD
#undef GET_ADJ_VF

	if(!EDITOR_HAS_CHANGES(e) && new_changes)
	{
	    e->has_changes = TRUE;
	    EditorUpdateTitle(e);
	}
	EditorViewDraw(e);
        EditorUpdateMenus(e);
}
