#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/fio.h"

#include "obj.h"
#include "sarfio.h"

#include "guiutils.h"
#include "cdialog.h"
#include "csd.h"

#include "editor.h"
#include "editorcb.h"
#include "editorobj.h"
#include "editorscnopen.h"
#include "editorlist.h"
#include "scnedit.h"
#include "config.h"


static gint EditorScenerySaveProgressCB(gpointer data, glong n, glong m);
gint EditorSceneryDoQuerySave(editor_struct *e);
static void EditorScenerySaveObject(
	editor_struct *e, void ***parm, int *total_parms,
	editor_obj_struct *o, gboolean verbose
);
static void EditorScenerySaveHeader(
	editor_struct *e, void ***parm, int *total_parms,
	gboolean verbose
);
void EditorScenerySave(
        editor_struct *e, const gchar *path, gboolean verbose
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? (gfloat)atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#define DEGTORAD(d)     ((d) * PI / 180)
#define RADTODEG(r)     ((r) * 180 / PI)


/*
 *	Save progress callback, this just updates the status bar.
 */
static gint EditorScenerySaveProgressCB(gpointer data, glong n, glong m)
{
        editor_struct *e = EDITOR(data);
        if((e == NULL) || (m <= 0))
            return(-1);

        if(EDITOR_STOP_COUNT(e))
            return(-1);

        StatusBarProgress(
            e->status_bar, (gfloat)n / (gfloat)m, TRUE
        );

        return(0);
}

/*
 *      Checks if there are changes on the editor, if there are then
 *      the user will be prompted to save the changes.
 *
 *      If there are changes and the user chooses yes and the
 *      save was successful, then 0 is returned.
 *
 *      If there are no changes, then 0 is returned.
 *
 *      If there are changes and the user chooses no, then 0 is
 *      returned.
 *
 *      If there are changes but the user chooses cancel or there
 *      was an error saving, then non-zero is returned.
 *
 *	In conclusion, the operation should not continue if a
 *	non-zero value was returned.
 */
gint EditorSceneryDoQuerySave(editor_struct *e)
{
        if(e == NULL)
            return(-1);

        if(EDITOR_HAS_CHANGES(e))
        {
            gint status;
            GtkWidget *toplevel = EDITOR_TOPLEVEL(e);

            CDialogSetTransientFor(toplevel);
            status = CDialogGetResponse(
                "Save Changes?",
"The scenery data contains changes that have not been\n\
saved, do you want to save those changes?",
"There are changes made to the scenery data that have\n\
not been saved. If you want to save those changes then\n\
click on `Yes'.  If you want to discard those changes\n\
then click on `No'.  If you do not want to proceed\n\
with this function then click on `Cancel'.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
                CDIALOG_BTNFLAG_CANCEL | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_YES
            );
            CDialogSetTransientFor(NULL);
            switch(status)
            {
              case CDIALOG_RESPONSE_YES_TO_ALL:
              case CDIALOG_RESPONSE_YES:
              case CDIALOG_RESPONSE_OK:
                /* Save scenery. */
                EditorSaveCB(NULL, e);
                /* If the scenery was successfully saved then there
                 * should not be any changes.
                 */
                if(EDITOR_HAS_CHANGES(e))
                    return(-1);
                else
                    return(0);
                break;

              case CDIALOG_RESPONSE_NO:
                return(0);
                break;

              default:  /* Assume cancel. */
                return(-1);
                break;
            }
        }
        else
        {
            return(0);
        }
}

/*
 *	Called by EditorScenerySave() to append parameters for saving
 *	of the given object.
 */
static void EditorScenerySaveObject(
        editor_struct *e, void ***parm, int *total_parms,
	editor_obj_struct *o, gboolean verbose
)
{
        void *p;
	const sar_contact_bounds_struct *cb = &o->cb;

#define NEW_PARM(t)     (SARParmNewAppend((t),parm,total_parms))

	/* Create object. */
	if(TRUE)
	{
	    /* Create static object? */
	    if(o->type == SAR_OBJ_TYPE_STATIC)
	    {
#define PARM_TYPE	sar_parm_new_object_struct
		PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
		    SAR_PARM_NEW_OBJECT
		);
		pv->object_type = SAR_OBJ_TYPE_STATIC;
#undef PARM_TYPE
	    }
	    /* Create helipad? */
	    else if(o->type == SAR_OBJ_TYPE_HELIPAD)
            {
#define PARM_TYPE	sar_parm_new_helipad_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                    SAR_PARM_NEW_HELIPAD
                );

                pv->style = STRDUP(o->helipad_type_name);
		pv->length = o->helipad_length;
		pv->width = o->helipad_width;
		pv->recession = o->helipad_recession;
		pv->label = STRDUP(o->helipad_label);
		pv->flags |= o->helipad_flags;

		pv->ref_obj_name = STRDUP(o->ref_obj_name);
		if(pv->ref_obj_name != NULL)
		{
		    pv->flags |=	SAR_HELIPAD_FLAG_REF_OBJECT |
					SAR_HELIPAD_FLAG_FOLLOW_REF_OBJECT;
		    memcpy(&pv->ref_offset, &o->ref_offset, sizeof(sar_position_struct));
		    memcpy(&pv->ref_dir, &o->ref_dir, sizeof(sar_direction_struct));
		}
#undef PARM_TYPE
            }
            /* Create runway? */
            else if(o->type == SAR_OBJ_TYPE_RUNWAY)
            {
#define PARM_TYPE       sar_parm_new_runway_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                    SAR_PARM_NEW_RUNWAY
                );

		pv->range = o->range;
		pv->length = o->runway_length;
		pv->width = o->runway_width;
		pv->surface = o->runway_surface;
		pv->dashes = o->runway_dashes;
		pv->edge_light_spacing = o->runway_edge_light_spacing;
		pv->north_label = STRDUP(o->runway_north_label);
		pv->south_label = STRDUP(o->runway_south_label);
		pv->north_displaced_threshold =
		    o->runway_north_displaced_threshold;
		pv->south_displaced_threshold =
                    o->runway_south_displaced_threshold;
		pv->flags = o->runway_flags;
#undef PARM_TYPE
	    }
            /* Create human? */
            else if(o->type == SAR_OBJ_TYPE_HUMAN)
            {
#define PARM_TYPE       sar_parm_new_human_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                    SAR_PARM_NEW_HUMAN
                );

		pv->type_name = STRDUP(o->human_type_name);
		pv->flags |= o->human_flags;
#undef PARM_TYPE
            }
            /* Create fire? */
            else if(o->type == SAR_OBJ_TYPE_FIRE)
            {
#define PARM_TYPE	sar_parm_new_fire_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                    SAR_PARM_NEW_FIRE
                );

		pv->radius = o->fire_radius;
		pv->height = o->fire_height;
#undef PARM_TYPE
            }
            /* Create smoke? */
            else if(o->type == SAR_OBJ_TYPE_SMOKE)
            {
#define PARM_TYPE       sar_parm_new_smoke_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                    SAR_PARM_NEW_SMOKE
                );

		memcpy(&pv->offset, &o->smoke_offset, sizeof(sar_position_struct));
		pv->radius_start = o->smoke_radius_start;
		pv->radius_max = o->smoke_radius_max;
		pv->radius_rate = o->smoke_radius_rate;
		pv->hide_at_max = o->smoke_hide_at_max;
		pv->respawn_int = o->smoke_respawn_int;
		pv->total_units = o->smoke_total_units;
		pv->color_code = o->smoke_color_code;
#undef PARM_TYPE
            }
            /* Create premodeled? */
            else if(o->type == SAR_OBJ_TYPE_PREMODELED)
            {
#define PARM_TYPE	sar_parm_new_premodeled_struct
                PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
		    SAR_PARM_NEW_PREMODELED
                );

		g_free(pv->model_type);
		pv->model_type = STRDUP(o->premodeled_type_name);
		pv->argv = strexp(o->premodeled_value, &pv->argc);
#undef PARM_TYPE
            }
	}
	/* Model file. */
        if((o->model_file != NULL) ? (*o->model_file != '\0') : FALSE)
        {
#define PARM_TYPE	sar_parm_model_file_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_MODEL_FILE
            );
	    g_free(pv->file);
	    pv->file = STRDUP(o->model_file);
#undef PARM_TYPE
        }
	/* Name. */
	if((o->name != NULL) ? (*o->name != '\0') : FALSE)
	{
#define PARM_TYPE	sar_parm_object_name_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_OBJECT_NAME
            );
	    g_free(pv->name);
	    pv->name = STRDUP(o->name);
#undef PARM_TYPE
        }
        /* Map description. */
        if((o->description != NULL) ? (*o->description != '\0') : FALSE)
        {
#define PARM_TYPE	sar_parm_object_map_description_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_OBJECT_MAP_DESCRIPTION
            );
            g_free(pv->description);
            pv->description = STRDUP(o->description);
#undef PARM_TYPE
        }
	/* Range. */
	if(o->range > 0.0f)
	{
#define PARM_TYPE       sar_parm_range_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_RANGE
            );
	    pv->range = o->range;
#undef PARM_TYPE
	}
        /* Range far. */
        if(o->range_far > 0.0f)
        {
#define PARM_TYPE       sar_parm_range_far_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_RANGE_FAR
            );
            pv->range_far = o->range_far;
#undef PARM_TYPE
        }
	/* Translate. */
	if(TRUE)
	{
#define PARM_TYPE	sar_parm_translate_struct
	    PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
		SAR_PARM_TRANSLATE
	    );
	    memcpy(&pv->translate, &o->pos, sizeof(sar_position_struct));
#undef PARM_TYPE
	}
        /* Rotate. */
        if(TRUE)
        {
#define PARM_TYPE       sar_parm_rotate_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_ROTATE
            );
            memcpy(&pv->rotate, &o->dir, sizeof(sar_direction_struct));
#undef PARM_TYPE
        }
        /* No depth test. */
        if(o->flags & SAR_OBJ_FLAG_NO_DEPTH_TEST)
        {
	    p = NEW_PARM(SAR_PARM_NO_DEPTH_TEST);
        }
        /* Polygon offset. */
        if((o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET) ||
           (o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_REVERSE) ||
           (o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_WRITE_DEPTH)
	)
        {
#define PARM_TYPE	sar_parm_polygon_offset_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
		SAR_PARM_POLYGON_OFFSET
            );
	    if(o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET)
		pv->flags |= SAR_OBJ_FLAG_POLYGON_OFFSET;
            if(o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_REVERSE)
                pv->flags |= SAR_OBJ_FLAG_POLYGON_OFFSET_REVERSE;
            if(o->flags & SAR_OBJ_FLAG_POLYGON_OFFSET_WRITE_DEPTH)
                pv->flags |= SAR_OBJ_FLAG_POLYGON_OFFSET_WRITE_DEPTH;
#undef PARM_TYPE
        }
	/* Contact bounds spherical. */
	if((cb->contact_shape == SAR_CONTACT_SHAPE_SPHERICAL) &&
	   (cb->contact_radius > 0.0)
	)
	{
#define PARM_TYPE       sar_parm_contact_bounds_spherical_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_CONTACT_BOUNDS_SPHERICAL
            );
	    pv->radius = cb->contact_radius;
#undef PARM_TYPE
	}
        /* Contact bounds cylendrical. */
        if((cb->contact_shape == SAR_CONTACT_SHAPE_CYLENDRICAL) &&
           (cb->contact_radius > 0.0)
        )
        {
#define PARM_TYPE       sar_parm_contact_bounds_cylendrical_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_CONTACT_BOUNDS_CYLENDRICAL
            );
            pv->radius = cb->contact_radius;
	    pv->height_min = cb->contact_h_min;
	    pv->height_max = cb->contact_h_max;
#undef PARM_TYPE
        }
        /* Contact bounds rectangular. */
        if((cb->contact_shape == SAR_CONTACT_SHAPE_RECTANGULAR) &&
           ((cb->contact_x_min != 0.0) || (cb->contact_x_max != 0.0))
        )
        {
#define PARM_TYPE       sar_parm_contact_bounds_rectangular_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_CONTACT_BOUNDS_RECTANGULAR
            );
            pv->x_min = cb->contact_x_min;
            pv->x_max = cb->contact_x_max;
            pv->y_min = cb->contact_y_min;
            pv->y_max = cb->contact_y_max;
            pv->z_min = cb->contact_z_min;
            pv->z_max = cb->contact_z_max;
#undef PARM_TYPE
        }
        /* Runway approach lighting north. */
        if((o->type == SAR_OBJ_TYPE_RUNWAY) &&
	   (o->runway_north_lighting_flags != 0)
        )
        {
#define PARM_TYPE       sar_parm_runway_approach_lighting_north_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_RUNWAY_APPROACH_LIGHTING_NORTH
            );
            pv->flags = o->runway_north_lighting_flags;
#undef PARM_TYPE
        }
        /* Runway approach lighting south. */
        if((o->type == SAR_OBJ_TYPE_RUNWAY) &&
           (o->runway_south_lighting_flags != 0)
        )
        {
#define PARM_TYPE       sar_parm_runway_approach_lighting_south_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_RUNWAY_APPROACH_LIGHTING_SOUTH
            );
            pv->flags = o->runway_south_lighting_flags;
#undef PARM_TYPE
        }
	/* Human message enter. */
	if((o->type == SAR_OBJ_TYPE_HUMAN) &&
           !IS_STRING_EMPTY(o->human_message_enter)
        )
	{
#define PARM_TYPE       sar_parm_human_message_enter_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_HUMAN_MESSAGE_ENTER
            );
	    g_free(pv->message);
	    pv->message = STRDUP(o->human_message_enter);
#undef PARM_TYPE
	}
	/* Human reference. */
	if((o->type == SAR_OBJ_TYPE_HUMAN) &&
	   !IS_STRING_EMPTY(o->ref_obj_name)
	)
	{
#define PARM_TYPE	sar_parm_human_reference_struct
	    PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_HUMAN_REFERENCE
            );

	    if(o->human_flags & SAR_HUMAN_FLAG_RUN_TOWARDS)
		pv->flags |= SAR_HUMAN_FLAG_RUN_TOWARDS;
	    else if(o->human_flags & SAR_HUMAN_FLAG_RUN_AWAY)
		pv->flags |= SAR_HUMAN_FLAG_RUN_AWAY;

	    g_free(pv->reference_name);
	    pv->reference_name = STRDUP(o->ref_obj_name);
#undef PARM_TYPE
	}
/* TODO */

#undef NEW_PARM
}

/*
 *	Called by EditorScenerySave() to append parameters for saving
 *	of the header.
 */
static void EditorScenerySaveHeader(
	editor_struct *e, void ***parm, int *total_parms,
	gboolean verbose
)
{
	void *p;


#define NEW_PARM(t)	(SARParmNewAppend((t),parm,total_parms))

	/* Name. */
	if(TRUE)
	{
#define PARM_TYPE	sar_parm_name_struct
	    PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
		SAR_PARM_NAME
	    );
	    GtkEntry *entry = (GtkEntry *)e->hdr_name_entry;
	    g_free(pv->name);
	    pv->name = STRDUP(
		(entry != NULL) ? gtk_entry_get_text(entry) : "Untitled"
	    );
#undef PARM_TYPE
	}
	/* Scene GPS. */
	if(TRUE)
	{
#define PARM_TYPE       sar_parm_scene_gps_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_GPS
            );
            GtkAdjustment *adj;
	    GtkSpinButton *spin;

	    spin = (GtkSpinButton *)e->hdr_gps_longitude_spin;
	    adj = (spin != NULL) ? spin->adjustment : NULL;
	    pv->dms_x_offset = (adj != NULL) ? adj->value : 0.0f;
            spin = (GtkSpinButton *)e->hdr_gps_latitude_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->dms_y_offset = (adj != NULL) ? adj->value : 0.0f;
            spin = (GtkSpinButton *)e->hdr_planetary_radius_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->planet_radius = (adj != NULL) ? adj->value : 6371000.0f;
#undef PARM_TYPE
	}
        /* Scene CANT. */
        if(TRUE)
        {
#define PARM_TYPE       sar_parm_scene_cant_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_CANT
            );
            GtkSpinButton *spin = (GtkSpinButton *)e->hdr_cant_spin;
	    GtkAdjustment *adj = (spin != NULL) ? spin->adjustment : NULL;
	    pv->cant = (gfloat)DEGTORAD((adj != NULL) ? adj->value : 0.0f);
#undef PARM_TYPE
	}
        /* Scene elevation. */
        if(TRUE)
        {
#define PARM_TYPE       sar_parm_scene_elevation_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_ELEVATION
            );
            GtkSpinButton *spin = (GtkSpinButton *)e->hdr_elevation_spin;
            GtkAdjustment *adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->elevation = (gfloat)SFMFeetToMeters(
		(adj != NULL) ? adj->value : 0.0f
	    );
#undef PARM_TYPE
        }
	/* Texture load. */
	if(TRUE)
	{
#define PARM_TYPE	sar_parm_texture_load_struct
	    editclist_struct *eclist = e->hdr_textures_eclist;
	    GtkCList *clist = (GtkCList *)EDITCLIST_CLIST(eclist);
	    if((clist != NULL) ? (clist->columns >= 3) : FALSE)
	    {
		gint row, column;
		gchar *s;
#define GET_CLIST_TEXT						\
{ guint8 spacing; GdkPixmap *pixmap; GdkBitmap *mask;		\
 s = NULL;							\
 switch((gint)gtk_clist_get_cell_type(clist, row, column)) {	\
 case GTK_CELL_TEXT:						\
  gtk_clist_get_text(clist, row, column, &s);			\
  break;							\
 case GTK_CELL_PIXTEXT:						\
  gtk_clist_get_pixtext(					\
   clist, row, column, &s,					\
   &spacing, &pixmap, &mask					\
  );								\
  break;							\
} }

		for(row = 0; row < clist->rows; row++)
		{
		    PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
			SAR_PARM_TEXTURE_LOAD
		    );
		    column = 0;
		    GET_CLIST_TEXT
		    g_free(pv->name);
		    pv->name = STRDUP(s);

		    column = 1;
		    GET_CLIST_TEXT
		    g_free(pv->file);
		    pv->file = STRDUP(s);

                    column = 2;
                    GET_CLIST_TEXT
                    pv->priority = (gfloat)CLIP(ATOF(s), 0.0, 1.0);
		}
#undef GET_CLIST_TEXT
	    }
#undef PARM_TYPE
	}
        /* Scene ground base flags. */
        if(TRUE)
        {
#define PARM_TYPE       sar_parm_scene_ground_flags_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_GROUND_FLAGS
            );
            GtkToggleButton *tb = (GtkToggleButton *)e->hdr_ground_is_water_tb;
	    if((tb != NULL) ? tb->active : FALSE)
		pv->flags |= SAR_SCENE_BASE_FLAG_IS_WATER;
#undef PARM_TYPE
        }
        /* Scene ground base tile. */
        if(TRUE)
        {
#define PARM_TYPE       sar_parm_scene_ground_tile_struct
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_GROUND_TILE
            );
	    csd_color_struct c;
	    GtkEntry *entry = (GtkEntry *)e->hdr_ground_texture_entry;
            GtkSpinButton *spin = (GtkSpinButton *)e->hdr_ground_tile_width_spin;
            GtkAdjustment *adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->tile_width = (gint)(
                (adj != NULL) ? adj->value : 0.0f
            );
	    spin = (GtkSpinButton *)e->hdr_ground_tile_height_spin;
	    adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->tile_height = (gint)(
                (adj != NULL) ? adj->value : 0.0f
            );
            spin = (GtkSpinButton *)e->hdr_ground_tile_close_range_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
            pv->close_range = (adj != NULL) ? adj->value : 0.0f;
	    CSDColorButtonGetColor(e->hdr_ground_color_btn, &c);
	    pv->color.a = 1.0;
	    pv->color.r = c.r;
            pv->color.g = c.g;
            pv->color.b = c.b;
	    g_free(pv->texture_name);
	    pv->texture_name = STRDUP(
		(entry != NULL) ? gtk_entry_get_text(entry) : NULL
	    );
#undef PARM_TYPE
        }
        /* Scene map. */
        if(TRUE)
        {
#define PARM_TYPE	sar_parm_scene_map_struct
	    GtkAdjustment *adj;
	    GtkSpinButton *spin;
	    editor_tex_struct *t = e->hdr_map_tex;
            PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                SAR_PARM_SCENE_MAP
            );
	    g_free(pv->file);
	    pv->file = STRDUP((t != NULL) ? t->filename : NULL);
            spin = (GtkSpinButton *)e->hdr_map_width_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
	    pv->width = (adj != NULL) ? adj->value : 1000.0f;
            spin = (GtkSpinButton *)e->hdr_map_height_spin;
            adj = (spin != NULL) ? spin->adjustment : NULL;
	    pv->height = (adj != NULL) ? adj->value : 1000.0f;
#undef PARM_TYPE
        }
        /* Registered location. */
        if(TRUE)
        {
#define PARM_TYPE	sar_parm_register_location_struct
            editclist_struct *eclist = e->hdr_registered_locations_eclist;
            GtkCList *clist = (GtkCList *)EDITCLIST_CLIST(eclist);
            if((clist != NULL) ? (clist->columns >= 7) : FALSE)
            {
                gint row, column;
                gchar *s;
#define GET_CLIST_TEXT						\
{ guint8 spacing; GdkPixmap *pixmap; GdkBitmap *mask;		\
 s = NULL;							\
 switch((gint)gtk_clist_get_cell_type(clist, row, column)) {	\
 case GTK_CELL_TEXT:						\
  gtk_clist_get_text(clist, row, column, &s);			\
  break;							\
 case GTK_CELL_PIXTEXT:                                         \
  gtk_clist_get_pixtext(                                        \
   clist, row, column, &s,                                      \
   &spacing, &pixmap, &mask                                     \
  );                                                            \
  break;                                                        \
} }
                for(row = 0; row < clist->rows; row++)
                {
                    PARM_TYPE *pv = (PARM_TYPE *)p = NEW_PARM(
                        SAR_PARM_REGISTER_LOCATION
                    );
                    column = 0;
                    GET_CLIST_TEXT
                    g_free(pv->name);
                    pv->name = STRDUP(s);

                    column = 1;
                    GET_CLIST_TEXT
		    pv->pos.x = ATOF(s);
                    column = 2;
                    GET_CLIST_TEXT
                    pv->pos.y = ATOF(s);
                    column = 3;
                    GET_CLIST_TEXT
                    pv->pos.z = (gfloat)SFMFeetToMeters(ATOF(s));

                    column = 4;
                    GET_CLIST_TEXT
                    pv->dir.heading = (gfloat)DEGTORAD(ATOF(s));
                    column = 5;
                    GET_CLIST_TEXT
                    pv->dir.pitch = (gfloat)DEGTORAD(ATOF(s));
                    column = 6;
                    GET_CLIST_TEXT
                    pv->dir.bank = (gfloat)DEGTORAD(ATOF(s));
                }
#undef GET_CLIST_TEXT
            }
#undef PARM_TYPE
        }



#undef NEW_PARM
}

/*
 *	Saves scenery data on editor to file.
 */
void EditorScenerySave(
        editor_struct *e, const gchar *path, gboolean verbose
)
{
	void **parm = NULL;
	gint total_parms = 0;
        GtkWidget *toplevel;
        GtkCList *clist;


#define STATUS_BAR_MESSAGE(m)   \
 StatusBarMessage(e->status_bar, (m), TRUE);


        if((path == NULL) || (e == NULL))
            return;

	if(EDITOR_PROCESSING(e))
	    return;

        toplevel = EDITOR_TOPLEVEL(e);
        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

if(TRUE) {
 gint status;
 CDialogSetTransientFor(toplevel);
 status = CDialogGetResponse(
"Save Warning",
"The EditorScenerySave() function is currently under construction.\n\
\n\
Are you sure you want to save this file?",
"This function is currently under construction, the file that you\n\
save may not contain all the data that the original file contained.\n\
\n\
It is recommended that you save to an alternate file name so that\n\
you do not overwrite the original file.",
  CDIALOG_ICON_WARNING,
  CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
  CDIALOG_BTNFLAG_HELP,
  CDIALOG_BTNFLAG_NO
 );
 CDialogSetTransientFor(NULL);
 if(status != CDIALOG_RESPONSE_YES)
  return;
}

        e->processing = TRUE;
        EditorStopReset(e, TRUE);

        if(verbose)
            STATUS_BAR_MESSAGE("Saving scenery to file...");

	/* Apply changes to currently selected object. */
	EditorObjListDoApplyValues(e, NULL);

        /* Realize all objects before saving. */
        EditorObjListRealizeAll(e, TRUE);

	/* Begin generating parameters for saving. */

	/* Header parameters. */
	if(EDITOR_STOP_COUNT(e) <= 0)
	{
            if(verbose)
                STATUS_BAR_MESSAGE("Encoding header...");

	    EditorScenerySaveHeader(
		e, &parm, &total_parms, verbose
	    );
	}

	/* Object parameters. */
	if(EDITOR_STOP_COUNT(e) <= 0)
	{
	    gint i;

            if(verbose)
                STATUS_BAR_MESSAGE("Encoding objects...");

	    /* Iterate through each object and save them. */
	    for(i = 0; i < clist->rows; i++)
	    {
		editor_obj_struct *o = EDITOR_OBJ(
		    gtk_clist_get_row_data(clist, i)
		);
		if(o == NULL)
		    continue;

		if(EditorScenerySaveProgressCB(e, i + 1, clist->rows))
		    break;

		EditorScenerySaveObject(
		    e, &parm, &total_parms,
		    o, verbose
		);
	    }
	}

	/* At this point the parameters have been generated, now we
	 * save them.
	 */
	if(EDITOR_STOP_COUNT(e) <= 0)
	{
	    if(verbose)
		STATUS_BAR_MESSAGE("Writing scenery file...");

	    /* Save parameters to scenery file. */
	    if(SARParmSaveToFile(
		path, SAR_FILE_FORMAT_SCENE,
		parm, total_parms,
		e, EditorScenerySaveProgressCB
	    ))
	    {
		/* Unable to save scenery. */
		if(verbose)
		{
		    gchar *buf = g_strdup_printf(
"Unable to save scenery file:\n\
\n\
    %s",
			path
		    );
		    CDialogSetTransientFor(toplevel);
		    CDialogGetResponse(
			"Save Failed",
			buf,
"Check if the file exists and does not permit writing, or\n\
the directory does not permit writing.  Also check if there\n\
is sufficient storage space.  The scenery data may also be\n\
corrupt or that you have given a bad value somewhere.",
			CDIALOG_ICON_ERROR,
			CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
			CDIALOG_BTNFLAG_OK
		    );
		    CDialogSetTransientFor(NULL);
		    g_free(buf);

		    STATUS_BAR_MESSAGE("Error saving scenery file");
		}
	    }
	    else
	    {
		/* Scenery saved successfully. */

		/* Reset has changes mark. */
		e->has_changes = FALSE;

                if(verbose)
                    STATUS_BAR_MESSAGE("Scenery file saved");
	    }
	}

	/* Delete all parameters. */
	SARParmDeleteAll(&parm, &total_parms);

	if(verbose)
	    StatusBarProgress(e->status_bar, 0.0f, TRUE);

        e->processing = FALSE;
        EditorStopReset(e, FALSE);

#undef STATUS_BAR_MESSAGE
}

