#include "../include/fio.h"
#include "../include/disk.h"
#include <gtk/gtk.h>
#include "options.h"
#include "config.h"

options_struct *OptionsNew(void);
void OptionsDelete(options_struct *o);

gint OptionsLoadFromFile(options_struct *o, const gchar *filename);
gint OptionsSaveToFile(const options_struct *o, const gchar *filename);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Creates new options.
 */
options_struct *OptionsNew(void)
{
	options_struct *o = OPTIONS(g_malloc0(
	    sizeof(options_struct)
	));
	if(o == NULL)
	    return(NULL);

	o->cfg_file = NULL;
	o->sar_dir = STRDUP(SE_DEF_SAR_DATA_DIR);
	o->editor_last_x = SE_DEF_EDITOR_X;
        o->editor_last_y = SE_DEF_EDITOR_Y;
        o->editor_last_width = SE_DEF_EDITOR_WIDTH;
        o->editor_last_height = SE_DEF_EDITOR_HEIGHT;
        o->editor_last_page = 0;
	o->editor_last_scenery_paned_position = 330;
	o->editor_last_view_paned_position = 180;
	o->editor_last_units = 0;
	o->last_scn_filename = NULL;

	return(o);
}

/*
 *	Deletes options.
 */
void OptionsDelete(options_struct *o)
{
	if(o == NULL)
	    return;

	g_free(o->cfg_file);
	g_free(o->sar_dir);
	g_free(o->last_scn_filename);
	g_free(o);
}


/*
 *	Loads options from file.
 */
gint OptionsLoadFromFile(options_struct *o, const gchar *filename)
{
	FILE *fp;
	gchar *buf = NULL;

        if((o == NULL) || (filename == NULL))
            return(-1);

        /* Record configuration file name. */
        g_free(o->cfg_file);
        o->cfg_file = STRDUP(filename);

	/* Open file for reading. */
        fp = FOpen(filename, "rb");
        if(fp == NULL)
            return(0);

        do
        {
            buf = FSeekNextParm(fp, buf, '#', '=');
            if(buf == NULL)
                break;

            /* Version. */
            if(!g_strcasecmp(buf, "Version"))
            {
#define t	3
		gint vi[t];
                FGetValuesI(fp, vi, t);
#undef t
            }
            /* EditorLastGeometry */
            else if(!g_strcasecmp(buf, "EditorLastGeometry"))
            {
#define t	8
                gint vi[t];
                FGetValuesI(fp, vi, t);
		o->editor_last_x = vi[0];
		o->editor_last_y = vi[1];
		o->editor_last_width = MAX(vi[2], 100);
		o->editor_last_height = MAX(vi[3], 70);
		o->editor_last_page = vi[4];
		o->editor_last_scenery_paned_position = MAX(vi[5], 0);
		o->editor_last_view_paned_position = MAX(vi[6], 0);
		o->editor_last_units = vi[7];
#undef t
            }
	    /* EditorViewDisplayFlags */
	    else if(!g_strcasecmp(buf, "EditorViewDisplayFlags"))
            {
#define t       1
                gint vi[t];
                FGetValuesI(fp, vi, t);
		o->editor_view_display_flags = (gulong)vi[0];
#undef t
            }
            /* EditorLastViewPosition */
            else if(!g_strcasecmp(buf, "EditorLastViewPosition"))
            {
#define t       6
                gdouble vf[t];
                FGetValuesF(fp, vf, t);
		o->editor_last_cursor_xw = (gint)vf[0];
		o->editor_last_cursor_yw = (gint)vf[1];
		o->editor_last_view_xm = (gfloat)vf[2];
		o->editor_last_view_ym = (gfloat)vf[3];
		o->editor_last_view_zoom = (gfloat)vf[4];
		o->editor_last_view_grid_spacing = (gfloat)vf[5];
#undef t
	    }
            /* LastSCNFile */
            else if(!g_strcasecmp(buf, "LastSCNFile"))
            {
		gchar *s = FGetStringLiteral(fp);
		if((s != NULL) ? (*s != '\0') : FALSE)
		{
		    g_free(o->last_scn_filename);
		    o->last_scn_filename = STRDUP(s);
		}
		g_free(s);
            }
            else
            {
                fprintf(
		    stderr,
                    "%s: Unsupported parameter `%s'.\n",
                    filename, buf
                );
                FSeekNextLine(fp);
            }

        } while(TRUE);

        FClose(fp);

	return(0);
}

/*
 *	Saves options to file.
 */
gint OptionsSaveToFile(const options_struct *o, const gchar *filename)
{
	const gchar *s;
	FILE *fp;
#ifndef __MSW__
	gchar *parent;
	struct stat stat_buf;
#endif

        if((o == NULL) || (filename == NULL))
            return(-1);

#ifndef __MSW__
	parent = STRDUP(GetParentDir(filename));
        if((parent != NULL) ? stat(parent, &stat_buf) : FALSE)
        {
            if(rmkdir(
		parent,
                S_IRUSR | S_IWUSR | S_IXUSR
            ))
                return(-1);
        }
#endif

        /* Open file for writing. */
        fp = FOpen(filename, "wb");
        if(fp == NULL)
            return(-1);

        /* Version */
        fprintf(
            fp,
            "Version = %i %i %i\n",
            PROG_VERSION_MAJOR, PROG_VERSION_MINOR, PROG_VERSION_RELEASE
        );
        /* EditorLastGeometry */
        fprintf(
            fp,
            "EditorLastGeometry = %i %i %i %i %i %i %i %i\n",
            o->editor_last_x, o->editor_last_y,
	    o->editor_last_width, o->editor_last_height,
            o->editor_last_page,
            o->editor_last_scenery_paned_position,
            o->editor_last_view_paned_position,
	    o->editor_last_units
        );
	/* EditorViewDisplayFlags */
	fprintf(
	    fp,
	    "EditorViewDisplayFlags = %ld\n",
	    o->editor_view_display_flags
	);
	/* EditorLastViewPosition */
	fprintf(
	    fp,
	    "EditorLastViewPosition = %i %i %f %f %f %f\n",
	    o->editor_last_cursor_xw,
	    o->editor_last_cursor_yw,
	    o->editor_last_view_xm,
	    o->editor_last_view_ym,
	    o->editor_last_view_zoom,
	    o->editor_last_view_grid_spacing
	);
        /* LastSCNFile. */
	s = o->last_scn_filename;
	if((s != NULL) ? (*s != '\0') : FALSE)
	    fprintf(
		fp,
		"LastSCNFile = %s\n",
		s
	    );

        FClose(fp);

#ifndef __MSW__
	g_free(parent);
#endif

	return(0);
}

